
#ifndef __VLBI_PROTO__
#define __VLBI_PROTO__

#include "TCPSocket.h"
#include "UDPSocket.h"
#include "FileSocket.h"

#include <string>
#include <iostream>
#include <vector>
#include <cstdio>
#include <cstdlib>
#include <algorithm>
using namespace std;

#include "VLBI_Frame.hpp"
#include "light_utl.hpp"
#include "netutl.h"
#include "Rate_meas.hpp"
#include "Trans_info.hpp"
#include "DataPage.hpp"
#include "ExtSubCh.hpp"

extern "C" {
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
}

const int DEFAULT_PORT = 26328;
const int DEFAULT_SNDR_PORT =    52100;
const int SNDBUFSIZE    =  16777216;
const int32_t BUF_SIZE_MAX  = 1000000000;
const int ERROR_CNT_LIMIT  =  10;

/** Class for VLBI data transmission prtocol
 *  abstract class
 */
class VLBI_Proto:public TransInfo{
public:
  /** Constructor */
  VLBI_Proto(const ApplProtocol_t apt=SUDP_Proto, 
	     const TransProtocol_t tpt=UDP_IO);
  /** Destructor  */
  virtual ~VLBI_Proto(){
    cerr << "VLBI_Proto:: Destructor runs\n";
    if(rate_meas.is_running()) rate_meas.stop();
    if(_sock !=0){ delete _sock; _sock=0;}
  }
  /// Creating Scoket
  int Create_sock();
  /// Setup VLBI frame 
  virtual void Set_VlbiFrame(VLBI_Frame * vf);
  virtual bool Tx_Ready(); ///< Check Readiness protocol.
  void Tx_Close(); ///< Closing connections
  bool Rx_Ready(const int port=0);  ///<**  Check/Make Rx Ready 
  /**  Transmitting data from Buffer for 1sec */
  virtual int32_t TX1sec(const time_t t, vector <uint8_t>& arr)=0;
  /// Tx 1sec of subset data  
  virtual int32_t
    TX1secSub(const time_t t, vector <uint8_t>& arr, const DAS_Info* indas=0)=0;
  /**  Transmitting data from Buffer for 1sec */
  virtual int32_t TX1sec(const time_t t, const uint8_t* arr, 
			 const int32_t ndat)=0;
  /// Tx 1sec of subset data  
  virtual int32_t
  TX1secSub(const time_t t, const uint8_t* arr, const int32_t ndat, 
	    const DAS_Info* indas=0)=0;
 
    //   /** @doc Receivind data to Buffer for 1sec */
    //   virtual int32_t RX1sec(vector <uint8_t>& arr)=0;
    /** @doc Transmit/write data to socket. */
    //  virtual  int32_t TX(const uint8_t* arr, const uint32_t nsize);
    /** @doc Receiving data to File */
    //  virtual int32_t RX(FILE *fp=NULL)=0;
    /**  Receiving UDP data stream */
  virtual int32_t RX_UDP(void* buf, const int32_t buf_size)=0;
  /// Receiving VDIF data stream
  virtual int RXvdif(DataPage& dpage)=0;
  /// Returns Head size of Application Procotol (e.g. SUDP)
  virtual int ApplProtoHeadSize()const=0;
  /// Rate measurement Start
  void Timer_Start(); 
  /// Rate measurement Stop
  void Timer_Stop(){ 
    rate_meas.stop(); 
    //    cerr << "Rate_Meas.Stop() commanded." << endl;
    //    rate_meas.join();
  }
  /** Compute Elapsed time from its Start  */ 
  double T_Elapsed();
  /** Print Rx data rate statistics */ 
  void Print_RX_rate(const int intvl_cnt, const int64_t seq_prev);
  //  string TransProtocol_Type();
  //  virtual ApplProtocol_t ApplProtocol_Type()const=0;
  /// Returns total amount of sent bytes in MBytes.
    double Sent_MBytes()const {
	return 1.e-6 * _sock->get_bytes_sent();} // (MB)
  /// Returns pointer of VLBI-Frame class
  VLBI_Frame* pVLBI_Frame()const{return vlbi_frame;}
  /// Tx Rate Adjustment by sleeping
  virtual void Rate_Adj(const int32_t sent_bytes=-1);
//   /// Rate Measurement start
//   void Rate_meas_start(){ if(!rate_meas.is_running()) rate_meas.start();}
//   /// Rate Measurement stop
//   void Rate_meas_stop(){if(rate_meas.is_running()) rate_meas.stop();}
  /// Set Tx Rate Display ON
  void Disp_TxRate(){rate_meas.Disp_Txon();}
  /// Set Rx Rate Display ON
  void Disp_RxRate(){rate_meas.Disp_Rxon();}
  /// Set Rate Display OFF
  void Disp_RateOff(){rate_meas.Disp_Off();}
  /// Set Rate Adjusting frequency
  void Set_Adj_Freq(const int n){
    rate_adj_frq = n; rate_meas.SetUpdateFreq((float)n);}
  /// Returns Rate Adjusting frequency
  int Adj_Freq()const {return rate_adj_frq;}
  /// Recording epoch and total byte sent at that time
  void Set_St1sec(){
      gettimeofday(&st_1sec,0); 
      tbyte_at1sec = _sock->get_bytes_sent();}
  /// Return Tx data rate including header (Mbps)
    double Tx_rate() const{return rate_meas.Tx_Mbps();}
  /// Return Rx data rate including header (Mbps)
  double Rx_rate() const{return rate_meas.Rx_Mbps();}
  /// Return net Tx data rate (Mbps)
  double Tx_pld_rate() const{return rate_meas.Tx_Pld_Mbps();}
  /// Return net Rx data rate (Mbps)
  double Rx_pld_rate() const{return rate_meas.Rx_Pld_Mbps();}
//   /** Extracting Sub Channel data from Data array, vector array version
//    * - Extracted data is stored in preSendBuf
//    * - Input DAS info is given from das
//    */
//   int   ExtSubCh(vector <uint8_t>& arr, const DAS_Info* indas);
//   /// Extracting Sub channels, pointer version
//   int   ExtSubCh(const uint8_t* arr, const int32_t ndat, const DAS_Info* indas);
  /** Set DAS Infor  
   * @return  0 for success and -1 for Error
   */
  int Set_DasInf(const DAS_Info& das);
  /** Set Extraction mode to Mk5VSI-geodsy */
  void SetExmode_Mk4Geo(){extch.Set_Mode(ExtSubCh::Mk4GeoVSI);}
protected:
  Socket *_sock;  ///< Socket for data transfer
  /** socket address structure used for UDP */
  struct sockaddr_in addr;
  /** Transmission Start time */
  struct timeval  st0;
  /// Total byte sent by the initial epoch(Bytes)
  uint64_t sent_bytes_initial; 
  /**  Start time at every 1sec data */
  struct timeval  st_1sec;
  uint64_t tbyte_at1sec; ///< Total Bytes at 1 sec tick
  /** Start time for rate adjustment */
  struct timeval  st_rate_adj;
  /** End time for rate adjustment */
  struct timeval  et_rate_adj;
  uint64_t sent_bytes_at; ///< Total bytes senet at that moment(MBytes)
  vector <uint8_t> packet; ///< Packet 
  /** Rate adjustment frequency (Hz), default 100 Hz */
  int rate_adj_frq;  
  VLBI_Frame *vlbi_frame;   ///< VLBI_Frame pointer
  //  TransInfo *trans_info;   ///< Transmission Information 
  Rate_Meas rate_meas; ///< Rate Measurement class
  /// Buffer to keep re-sampled data from input buf;
  //  vector <uint8_t> preSendBuf;
  ExtSubCh extch; ///< Data Extracion module
private:
  int32_t ComputePktSize(const int32_t frsize=0); // Computing Optimum Packet/frame size
  VLBI_Proto(VLBI_Proto& vp){;} ///< Copy Constructor, prohibited.
};

inline void VLBI_Proto::Rate_Adj(const int32_t sent_bytes){
  
    int64_t total_Bytes_sent =  _sock->get_bytes_sent();
    int64_t bdif = total_Bytes_sent - tbyte_at1sec;
    int64_t bdif0 = total_Bytes_sent - sent_bytes_initial;

    //   *(Logsystem::debug_logger()) << "Rate_Adj(): curr_sent_byte=" 
// 			       << current_MBytes_sent ;
//   *(Logsystem::debug_logger()) << " diff==" << bdif 
// 			     << " Adj=" <<Adj_Mbytes << endl;
    //  if( bdif < Adj_Mbytes) return 1;;
    
    gettimeofday(&et_rate_adj,0);   
    double byte = sent_bytes == -1 ? bdif0 : sent_bytes;
    double tdif = (double)(et_rate_adj.tv_sec - st_1sec.tv_sec)
	+1.e-6*(et_rate_adj.tv_usec-st_1sec.tv_usec);
    double tdif0 = (double)(et_rate_adj.tv_sec - st0.tv_sec)
	+1.e-6*(et_rate_adj.tv_usec-st0.tv_usec);

    double wait= 0.25 * byte / speed - tdif0;
    double inst_rate = 8.e-6*bdif/tdif;
    double avg_rate = 8.e-6*bdif0/tdif0;

    Logsystem::info_logger()->Tstamp()
      << "Rate_adj(): Inst_rate/Avg_rate =" << inst_rate 
      << "/" << avg_rate << "Mbps (target="<< speed*3.2e-5  << " Mbps),"
      << " Sleep=" << wait  << " sec\n";
    //  fprintf(stderr," Rate_adj(): Inst rate/avg_rate = %.1lf/%.1lf Mbps, (target: %5.1lf Mbps), sleep=%.5lf sec\n",
    //	  inst_rate, avg_rate, 3.2e-5*speed, wait);
    if(wait>0.0001){
	struct timespec dt={0, (uint32_t)(1000000000*wait)};
	nanosleep(&dt,0);
	//    usleep((uint32_t)(1000000*wait));
	//    fprintf(stderr," Rate_adj(): NanoSleep %e sec\n",wait);
	//     Logsystem::debug_logger()->Tstamp() << "byte/sec(" <<  byte << "/"
// 					<< tdif << ") = "
// 					<< inst_rate << "Mbps";
//     *Logsystem::debug_logger() << " Sleep=" << wait << endl;
    }
//   gettimeofday(&et,NULL);
//   tdif = (et.tv_sec-st0.tv_sec)+0.000001*(et.tv_usec-st0.tv_usec);
//   tx_rate_measured = 0.000001*(8.*byte)/tdif;  // (Mbps)
  
//   sent_bytes_at = current_MBytes_sent;  
//   st_rate_adj = et;

//  return wait;
}


inline  double VLBI_Proto::T_Elapsed(){
    struct timeval et_i, _diff;
    gettimeofday(&et_i,NULL); 
    timersub(&et_i, &st0, &_diff);
    double ret=_diff.tv_sec+_diff.tv_usec*1e-6;
    return(ret);
} 

inline void VLBI_Proto::Print_RX_rate(const int intvl_cnt, int64_t seq_prev){
    //  gettimeofday(&et,NULL);   
    double dt0 = rate_meas.T_Elapsed();
    float avg_rate = (float)(rate_meas.Total_rcv()*8.*1.e-6/dt0);
    string epoch = rate_meas.Epoch();
    float rate = rate_meas.Rx_Mbps();
    //  strftime(time_str,sizeof(time_str),"%Y/%m/%d-%H:%M:%S",localtime(&et.tv_sec));      
    fprintf(stderr,"[%s] Rate:%5.2f Mbps %6d/%6lld pkts",  
	    epoch.c_str(), rate, intvl_cnt, (long long)seq_prev);
    fprintf(stderr," Total(%.2lf MByte)/%.2lf sec = %5.2lf Mbps\n",
	    1.e-6*rate_meas.Total_rcv(), dt0, avg_rate);
    //  fprintf(stderr," Dup:%3d Miss:%3d \n", ndup, n_mis);
    //  st_i=et; 
}

//********************************* 
inline vector <string> StrSep(const char * s1, const char *sep){
    string  work =s1;
    size_t pos=0;
    vector <string> arr; 
    while( (pos=work.find(sep)) != work.npos){
	arr.push_back(work.substr(0,pos));
	work = work.substr(pos+1);
    };
    arr.push_back(work);  
  
    return arr;
}

inline vector <int> IntSep(const char * s1, const char *sep){
    string  work =s1;
    size_t pos=0;
    int t;
    vector <int> arr; 
    while( (pos=work.find(sep)) != work.npos){
	t = atoi(work.substr(0,pos).c_str());
	arr.push_back(t);
	work = work.substr(pos+1);
    };
    t = atoi(work.c_str());
    arr.push_back(t);  
  
    return arr;
}

#if 0
inline int32_t VLBI_Proto::TX(const uint8_t* arr, const uint32_t nsize){
    int32_t nsent;
    nsent = (int32_t)_sock->send((u_int8_t*)arr, nsize);
    if( nsent <0 ){
	cerr << "Error in VLBI_Proto::TX(): failed to write data " << nsize
	     << " bytes\n";
    }
    return nsent;
}
#endif


#endif // end of ifdef 
