
#ifndef __VSIDATA_HPP__
#define __VSIDATA_HPP__
#include <iostream>
#include <vector>
using namespace std;
#include "DAS_info.hpp"

extern "C"{
#include <stdint.h>
}

const int MAX_NBIT = 8;
const int MAX_NCH = 8;

/**
 * A Class for VSI data Handling
 */

class VsiData{
public:
  /** Constructor */
  VsiData():dasInfo(0){;}
  /** Make up look-up table. */
  const int Make_Table();
  /** Set DAS information class */
  void Set_DasInfo( DAS_Info* das){ dasInfo = das;  }
  /** Reading 1 sec of data from the file */
  const int32_t Read1Sec(FILE *fp);
  /** Value of the data at (ch,ith) */
  const int8_t Val(const int ch, const int32_t i);
  /** Interpretation (decoding) Processing */
  void Interpret();
  /** Retuins reference of value array at ch*/
  vector <int8_t>& ValArr(const int ch){ return valarr[ch];}
  /** Retuins pointer of value array at ch */
  const int8_t *pValarr(const int ch) {return &valarr[ch][0];} ///< Returns pointer of one
private:
  const int MkTbl1bit();
  const int MkTbl2bit();
  const int MkTbl4bit();
  const int MkTbl8bit();
  DAS_Info *dasInfo;
  vector <uint8_t> buf;
  vector <int8_t> valarr[16]; 
  int8_t table[256][MAX_NCH];
};

const int8_t level_1bit[2] = {-1,1};
const int8_t level_2bit[4] = {-3, -1, 1, 3 };


inline void VsiData::Interpret(){
  if(dasInfo ==0){  
    cerr << "Error @VsiData::Val():dasInfo is empty.\n";
    return ;  
  }
  
  int nch= dasInfo->Nch();
  int bit = dasInfo->Qbit();
  int dunit = nch * bit;
  int ch_byte = 8/bit;
  int unit_byte(0);
  int byte_unit (0);
  char msg[256];
  try{
    if( nch > ch_byte){  // 1 unit takes more than 1 ch
      if( dunit % 8 != 0){
	sprintf(msg, 
		"Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }
      byte_unit = dunit / 8;
    }else{  // 1 byte include no less than 1 channel
      if( 8 % dunit != 0 ){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }else{
	unit_byte = 8 / dunit;
      }
    }
  } 
  catch(const char* msg1){
    cerr << "Error in VsiData::Val():" << msg1 << endl;
    return ;
  }
  
  int32_t ndat = dasInfo->Srate();
  for( int i=0; i<nch; i++) 
    if( (int32_t)valarr[i].size() < ndat) valarr[i].resize(ndat);
  vector <uint8_t>::iterator sptr = buf.begin();  
  vector <int8_t>::iterator dptr[nch];
  for(int i=0; i<nch; i++) dptr[i] = valarr[i].begin();  

  int32_t cnt=0;
  
  if( byte_unit == 0){ // 1 byte contains multiple units
    while(sptr < buf.end() && dptr[0] < valarr[0].end()){
      for( int i=0; i< unit_byte; i++){
	for(int j=0; j<nch; j++){
	  *dptr[j] = table[*sptr][j+i*nch]; 
	  dptr[j]++;
	}
	cnt++;
      }
      sptr++;
    } 
  }else{ // 1 unit takes multiple bytes
    while(sptr < buf.end() && dptr[0] < valarr[0].end()){
      int  ich=0;
      for(int i=0; i<byte_unit; i++){ // bytes/unit
	for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
	  *dptr[ich] = table[*sptr][jch];  
	  dptr[jch]++;
	}
	sptr++;
      }
      cnt++;
    }
  } 
  
  return ;
}



inline const int8_t VsiData::Val(const int ch, const int32_t ith){
  if(dasInfo ==0){  
    cerr << "Error @VsiData::Val():dasInfo is empty.\n";
    return -100;  
  }
  int nch= dasInfo->Nch();
  int bit = dasInfo->Qbit();
  int dunit = nch * bit;
  int ch_byte = 8/bit;
  int unit_byte(0);
  int byte_unit (0);
  char msg[256];

  try{
    if( nch > ch_byte){  // 1 unit takes more than 1 ch
      if( dunit % 8 != 0){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }
      byte_unit = dunit / 8;
    }else{  // 1 byte include no less than 1 channel
      if(8 % dunit != 0){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }else{
	unit_byte = 8/dunit;
      }
    }
  } 
  catch(const char* msg){
    cerr << "Error in VsiData::Val():" << msg << endl;
    return -100;
  }

  if( byte_unit == 0 ){
    int32_t ibyte = ith / unit_byte;
    int frac_unit = ith %  unit_byte;
    return table[buf[ibyte]][ch+frac_unit*nch];
  }else{
    int32_t ibyte = ith*byte_unit;
    int abyte = ch /ch_byte;
    ibyte += abyte;
    int ich = ch % ch_byte;
    return table[buf[ibyte]][ich];
  }
  
}

#endif
