/**
 * @file  tds_file.cpp
 * @brief Implementation of TDS sampler (K5/VSSP) File data handling
 *
 * @author M.Sekido (NICT Kashima.)
 * @date $Date: 2010/01/02 03:40:37 $
 * @version $Id: tds_file.cpp,v 1.2 2010/01/02 03:40:37 sekido Exp $
 * 
 * Copyright (C) NICT Kashima. All rights reserved.
 */


#include  "tds_file.hpp"
#include  <algorithm>
#include  <cstring>
using namespace std;

extern "C" {
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
}

int  TDS_FILE::HD_Find(unsigned char *hd_dat){
  //  unsigned char ch;
  int vssp32_hd_size=32; // Byte
  int vssp_hd_size=8; // Byte

  unsigned char hd_buf[vssp32_hd_size];
  uint32_t *pb = (uint32_t *)hd_buf;
  int nread(0);

  while( ! feof(fp)){
    nread = fread((void*) hd_buf, 1, vssp32_hd_size, fp);
    if( nread < (int)vssp32_hd_size ){
      fprintf( stderr,"Error: sync pattern not found \n");
      return(HD_DETECT_FAILURE);
    }
    if( *pb == 0xffffffff )
      //      fprintf( stderr,"Sync pattern found!\n");
      switch(hd_buf[7]){
      case 0x8b:  // VSSP headder 
	strncpy((char *)hd_dat, (char *)hd_buf+4, vssp_hd_size -4);
	fseek(fp,-vssp32_hd_size,SEEK_CUR);
	return(HD_DETECT_SUCCESS);
	break;
      case 0x8c:  // VSSP32 headder 
	//	strncpy((char *)hd_dat, (char *)hd_buf+4, vssp32_hd_size -4);
	memcpy((char *)hd_dat, (char *)(hd_buf+4), (vssp32_hd_size -4));
	fseek(fp,-vssp32_hd_size,SEEK_CUR);
	return(HD_DETECT_SUCCESS);
	break;
      default:
	fseek(fp,1-vssp32_hd_size,SEEK_CUR);
      //      read_byte ++;
    }
  } // end while 
  return(FILE_END);
}

int  TDS_FILE::Initial_Check(const char *file){
  fname = file;
  
  struct stat fst;
  int ret(0);
  if( fp != NULL) fclose(fp);
  if( (ret = stat(file, &fst)) <0 || (fp= fopen(file, "r")) == NULL){
    fprintf(stderr, "Error: File(%s) cannot be open:%s\n", file,strerror(errno));
    return(-1);
  }

  TDS_HD hd_tmp;
  unsigned char hd_buf[28];
//   if( HD_Find(hd_buf) < 0){
//     fprintf(stderr, "Error: Could not find headder in the file \n");
//     exit(1);
//   }
  switch( HD_Find(hd_buf)){
  case HD_DETECT_FAILURE:
    fprintf(stderr, "Error: Could not find headder in the file \n");
    return -1;
    break;
  case FILE_END:
    fprintf(stderr, " FILE-END found in HD_Find.\n");
    return 0;
    break;
  }
  
  hd_tmp.Set_Data(hd_buf); // set head data
  if(tds_hd.size() < 1)  tds_hd.push_back(hd_tmp);
  else       tds_hd[0] = hd_tmp;

  double soy;

  switch(hd_tmp.SyncB()){
  case 0x8b:
    soy = (t_cur.Doy()-1)*86400 + hd_tmp.Sod();
    t0.Set_time(t_cur.Year(), soy);
    break;
  case 0x8c:
    t0.Set_time(hd_tmp.Year(), hd_tmp.Doy()*86400+hd_tmp.Sod());
    break;
  default:
    cerr << "Error in TDS_FILE::Initial Chech(): sync patter is not expected."
	 << endl;
    fprintf(stderr,"pattern= %x\n", hd_tmp.SyncB());
    return -1;
    //    exit(1);
  }

  Update_info(0); // Update all information with latest headder   
  Set_time_elapsed(0);//0 sec from t0
  total_byte_sec = sfrq /8 * nch * sbit;
  Set_buf(total_byte_sec+hbyte);
  
  return 1;
}

int  TDS_FILE::Skip_process(void){
  int32_t skip_pt=0;
  uint32_t skip_byte=0;
  
  if( skip_time !=0.0){
    Add_time_elapsed(skip_time);
    if( skip_time >= 1.0){ 
      int skip_sec = (int) skip_time;
      skip_byte =  (hbyte + total_byte_sec)* skip_sec;
      //        unsigned char hd_buf[4];
      //        if(HD_Find(hd_buf) <0){
      //  	fprintf(stderr,"No more Headder was not found.\n");
      //  	return(-1);
      //        }
      //        if(fseek(fp, skip_byte, SEEK_CUR) < 0){
      //  	fprintf(stderr,"Error: Filed to skip data \n");
      //  	return(-1);
      //        }
      bp = data_buf.begin() + skip_byte;
      
      skip_time -= skip_sec;
    }
    //        if(HD_Find(hd_buf) <0){
    //  	fprintf(stderr,"No more Headder was not found.\n");
    //  	return(-1);
    //        }
    //        excpt_byte=0;
    //        TDS_HD tmp_hd;
    //        unsigned int prv_sod = (tds_hd.end()-1)->Sod();
    //        tmp_hd.Set_Data(hd_buf,prv_sod);
    //        tds_hd.push_back(tmp_hd);
    //      }
    //      Update_info(1);// sampling param and t_cur is updated with headder data
    
    skip_pt = (uint32_t )(skip_time * sfrq);
    skip_byte = skip_pt/8 * sbit * nch;
//      if(fseek(fp, skip_byte, SEEK_CUR) < 0){
//        fprintf(stderr,"Error: Filed to skip data \n");
//        return(-1);
//      }
    bp += skip_byte;
    excpt_byte += skip_byte;
  } 
  skip_time =0.0;
  //  skip_byte=0;
  return(1);
}

//--------------------------------------------
const int TDS_FILE::Open(const char *file){
  
  int ret;
  if( (ret = TDS_FILE::Initial_Check(file)) <0) 
    return(ret);

  ndata = (unsigned int )(Sfrq() * t_span + 0.5);

  for(int i=0; i<nch; i++){
    farr[i].resize(ndata + n_extra);
    pfye[i] = farr[i].begin();
  }
  xarr.resize(ndata);
  //  pdxs = 
  pdxe = xarr.begin();
  Set_Valtbl();
  return 1;
}


void TDS_FILE::Set_tspan(const double msec, const bool init_farr_flag){
  t_span = msec*1.e-3;
  uint32_t tmp_ndat = uint32_t(msec * 1.e-3 * Sfrq());
  if(ndata != tmp_ndat){
    ndata = tmp_ndat;
    Reset_rewind_flag();
  }

  if( init_farr_flag){
    for(int i=0; i<nch; i++){
      if(farr[i].size() !=ndata){
	farr[i].resize(ndata+n_extra);
	pfye[i] = farr[i].begin();
      }
    }
    if(xarr.size() != ndata){
      xarr.resize(ndata);
      pdxe = xarr.begin();
    }
  }
  
}



const int32_t TDS_FILE::Get_data(char flag ){
  int32_t n_request = ndata;
  
  if(flag && !flag_NoInterpret){ // Updating iterators to refrect used data number
    int32_t n_remain = pfye[0] - farr[0].begin();
    if( n_remain < (int32_t)ndata)
      fprintf(stderr,"pfye[0]-farr[0].begin() =%d < ndata (%d)\n",
	      n_remain, ndata);
    int32_t nx_remain = pdxe - xarr.begin();
    if( nx_remain < (int32_t)ndata)
      fprintf(stderr,"pdxe-xarr.begin() =%d < ndata (%d)\n",
	      nx_remain, ndata);
    assert( pfye[0] - ndata >= farr[0].begin() &&
	   pdxe -ndata >= xarr.begin());
    for(int i=0; i<nch;i++)  pfye[i] -= ndata;	 
    pdxe -= ndata;
  }

  double requested_byte = ndata*nch*sbit/8;
  uint32_t n_ret =0;
  uint32_t remain = Buf_Remain();
  int ret(0);
  if(remain < requested_byte){ // Need more data
    uint32_t avlbl_pt = remain*8/nch/sbit;
    assert( avlbl_pt*nch*sbit/8 == remain);
    if(! flag_NoInterpret)
      n_ret = Interprit_data(avlbl_pt);
    n_request = ndata-avlbl_pt;
    ret = Read2buf(); 
    switch(ret){
    case FILE_END:
      cerr << "File end\n";    return 0;   break;
    case HD_DETECT_FAILURE:
      cerr << "Error at TDS_FILE::Get_data(): Read2buf() failed\n";
      return -1;   break;
    }
    
    Skip_hbyte(); // Skip header byte 
    if(Skip_process() <0){
      fprintf(stderr,"Error: Failed at Skip_process in Get_data\n");
      return(-1);
    }
  }
  //  assert(farr[0].end() > pfye[0] + n_request-1);
  if(! flag_NoInterpret)
    n_ret += Interprit_data(n_request);
  return n_ret;
}

const int32_t TDS_FILE::Get_rawdata(char* buf, const uint32_t nbyte ){
  //  int32_t n_request = ndata;

  
//   if(flag){ // Updating iterators to refrect used data number
//     if(pfye[0] - farr[0].begin() < ndata)
//       fprintf(stderr,"pfye[0]-farr[0].begin() =%d < ndata (%d)\n",
// 	      pfye[0] - farr[0].begin(), ndata);
//     if(pdxe - xarr.begin() < ndata)
//       fprintf(stderr,"pdxe-xarr.begin() =%d < ndata (%d)\n",
// 	      pdxe - xarr.begin(), ndata);
//     assert(pfye[0] - ndata >= farr[0].begin() &&
// 	   //	   fremain > ndata &&  
// 	   pdxe -ndata >= xarr.begin());
//     for(int i=0; i<nch;i++)  pfye[i] -= ndata;	 
//     //    fremain -= ndata;
//     pdxe -= ndata;
//   }

//  double requested_byte = ndata*nch*sbit/8;
  int32_t requested_byte = nbyte;
  uint32_t remain = Buf_Remain();
  char* pntr = buf;
  vector <unsigned char>::iterator &buf_pntr = Buffer_pointer();
  
  if( (int32_t)remain < requested_byte){ // Need more data
    memcpy(pntr,&buf_pntr[0], remain); 
    buf_pntr += remain;
    pntr += remain;
    requested_byte -= remain;

    int ret = Read2buf();
    if( ret == FILE_END){
      cerr << "File end" << endl;
      //      return ret;
    }
    if(ret < 1) return ret;

    Skip_hbyte(); // Skip header byte 
    if(Skip_process() <0){
      fprintf(stderr,"Error: Failed at Skip_process in Get_data\n");
      return(-1);
    }
  }

  buf_pntr = Buffer_pointer();
  memcpy(pntr,&buf_pntr[0], requested_byte);
  buf_pntr += requested_byte;
  return 0;
}

const int32_t TDS_FILE::Get_rawdata( vector <uint8_t>& buf ){

  int32_t requested_byte = buf.size();
  uint32_t remain = Buf_Remain();
  vector <unsigned char>::iterator dest_pntr = buf.begin();
  vector <unsigned char>::iterator &src_pntr = Buffer_pointer();
  
  if( (int32_t)remain < requested_byte){ // Need more data
    copy(src_pntr, src_pntr + remain, dest_pntr);
    //    memcpy(pntr,&buf_pntr[0], remain); 
    src_pntr += remain;
    dest_pntr += remain;
    requested_byte -= remain;

    if( Read2buf() == FILE_END){
      cerr << "File end\n";
      return FILE_END;
    }
    Skip_hbyte(); // Skip header byte 
    if(Skip_process() <0){
      fprintf(stderr,"Error: Failed at Skip_process in Get_data\n");
      return(-1);
    }
  }

  src_pntr = Buffer_pointer();
  copy(src_pntr, src_pntr + requested_byte, dest_pntr);
  //  memcpy(pntr,&buf_pntr[0], requested_byte);
  src_pntr += requested_byte;
  return requested_byte;
}
 
int TDS_FILE::Read2buf(void){

  TDS_HD   tmp_hd;

  //  unsigned char hd_dat[4];
  unsigned char hd_dat[28];   // VSSP32
  int iret = HD_Find(hd_dat);
  switch(iret){
  case HD_DETECT_FAILURE:
    cerr << "Error at HD_Find(): Failed to find" << endl; 
    return HD_DETECT_FAILURE;      break;
  case FILE_END:
    cerr << " END of File found in HD_Find()" << endl; 
    return FILE_END;      break;
  }
//   if( (iret = HD_Find(hd_dat)) != HD_DETECT_SUCCESS)
//     //    if(iret< 0) exit(EXIT_FAILURE);
//     if(iret< 0) return -1;
  
    unsigned int prv_sod = Current_hd().Sod();
    iret = tmp_hd.Set_Data(hd_dat,prv_sod);  
    if( iret <0) return iret;
    tds_hd.push_back(tmp_hd);
    

    if( Eof())       return(FILE_END); 
    //    uint32_t expected_byte = total_byte_sec - excpt_byte;
    uint32_t expected_byte = total_byte_sec +hbyte;
    assert( expected_byte <= data_buf.size());
    if( fread((void *)&data_buf[0], expected_byte, 1,fp) < 1 ){
      if(feof(fp))  return(FILE_END);
      else{
	cerr << "fread() failed in Read2buf() by unknown reason\n";
	//	exit(-1);
	return -1;
      }
    };
    
#ifdef SOLARIS
    remap_data(&buf[0], expected_byte);
#endif

    excpt_byte =0;
    bp = data_buf.begin();
    bpe = bp + expected_byte;

    //    Add_time_end( (double)excpt_byte/(double)total_byte_sec) ;  // Soy end of last data
    
    return(NORMAL);
}

/*
 * $Log: tds_samplr.hpp
 *  Author: $Author: sekido $
 *  Date: $Date: 2010/01/02 03:40:37 $
 *  $Header: /vsspfs1/work3/cvs-archive/VDIF/tds_file.cpp,v 1.2 2010/01/02 03:40:37 sekido Exp $
 *  Revision: $Revision: 1.2 $
 *  Source: $Source: /vsspfs1/work3/cvs-archive/VDIF/tds_file.cpp,v $
 */
