/**
 * @file  tds_file.hpp
 * @brief TDS sampler (K5/VSSP) File data handling class 
 *
 * @author M.Sekido (NICT Kashima.)
 * @date $Date: 2010/01/02 03:40:37 $
 * @version $Id: tds_file.hpp,v 1.2 2010/01/02 03:40:37 sekido Exp $
 * $Log: tds_file.hpp,v $
 * Revision 1.2  2010/01/02 03:40:37  sekido
 * Replace FirstBranch to main
 *
 * Revision 1.1.2.7  2009/10/08 14:22:18  sekido
 * DOXY comment debugging finished
 *
 * Revision 1.1.2.6  2009/09/28 10:32:07  sekido
 * sudp_recv2 compile.
 *
 * Revision 1.1.2.5  2009/09/18 04:50:07  sekido
 *  sudp_send has almost begain to work normally.
 *
 * Revision 1.1.2.4  2009/09/11 10:03:07  sekido
 *  Adjusting sudp_send...
 *
 * Revision 1.1.2.3  2009/09/10 12:55:28  sekido
 * Adjusting for sending out  VDIF/UDP packet.
 *
 * Revision 1.1.2.2  2009/09/09 11:54:33  vlbi
 * sudp_send runnig version is ready.
 *
 * Revision 1.1.2.1  2009/09/09 04:31:15  vlbi
 * sudp_send created and compiled.
 *
 * Revision 1.2  2009/08/26 08:59:53  sekido
 * Creating statics_vssp32.cpp for statistical amplitude analysis.
 *
 * Revision 1.1  2009/05/28 12:19:11  sekido
 * Registering appropriate selected source files.
 *
 * Revision 1.3  2007/08/31 09:59:47  sekido
 * Adopting to VSSP32.
 *
 * Revision 1.2  2007/05/14 05:59:59  sekido
 *  -- Append mode added for file output
 *  -- Doxygen comment added
 *  -- cvs keyword added
 *
 * Revision: $Revision: 1.2 $
 * Copyright (C) NICT Kashima. All rights reserved.
 */


#ifndef __TDS_FILE__
#define __TDS_FILE__

#include <vector>
#include <string>
#include <iostream>
using namespace std;

#include "tds_cls.hpp"
#include "time_lib.h"
#include <assert.h>
// using namespace tds;


class TDS_FILE:public TDS_DAT {
public:
  /** Constructor */
  TDS_FILE():time_elapsed(0),fp(0),skip_time(0),total_byte_sec(0),excpt_byte(0),
	     t_span(1.e-2), ndata(0),flag_rewind(0), flag_NoInterpret(false),
	      n_extra(0){
    t_cur.Set_time_to_now();
  }
  /** Destructor */
  ~TDS_FILE(void){ if(fp != 0) fclose(fp); fp=0;}
  /** Open data file */
  const int Open(const char* file); 
  /** Read data to internal buffer 
   * @param msec : length of data. If it data amount 
   *   is different from that set by Set_tspan(), buffer size changed.
   */ 
  const int32_t Read(const int msec=100);
  /** main function to read data to buffer */
  const int32_t Get_data(const char flag );
  /** Copying raw data from inner buffer to a user buffer */
  const int32_t Get_rawdata(char* buf,const uint32_t nbyte );
  /** Get Raw data to the buffer */
  const int32_t Get_rawdata( vector <uint8_t>& buf );
  /** Set time-span for one-time reading data amount */
  void Set_tspan(const double T, const bool init_farr_flag=true);
  /** Updating DAS information from current header contents. */
  const int Update_info(const int level=0);
  /** Returns reference of internal float vector */
  const vector <float>& Farr(const int ich=0) const{return farr[ich];}
  /** Returns pointer of internal float array */
  const float * pFarr(const int ich=0)const {return &farr[ich][0];}
  /** Reteurns time tag (Soy) data of the specified data */
  const double Xarr(const int ix){
    if( ix >= 0 && ix < (int)xarr.size())   return xarr[ix]; 
    else  return -1;
  }
  /** Set reading skip time */
  void Set_Tskip(float tskip){ skip_time = tskip; }
  /** Set year and doy legacy vssp data */
  void Set_yydoy(int yy, int doy){ t_cur.Set_time(yy,doy,0,0,0);}
  /**  Initial Check and setup DAS info/Time from header info.*/
  int  Initial_Check(const char *file);
  /** Return Number of data in one-time reading */
  const uint32_t Ndat(void) const {return ndata;}
  /** Return Total bytes in a second*/
  uint32_t Total_byte_sec(void){return(total_byte_sec);}
  /** End of file testing */
  bool Eof(void){ return (feof(fp) != 0);}
  void Disp_hd(FILE *fp = stdout); ///< Display header info
  /** Returns current header reference */
  TDS_HD & Current_hd(void) {return *(tds_hd.end()-1);}
  /** Returns specified header reference */
  TDS_HD &Tds_hd(int i){ 
    if( i < (int)tds_hd.size()) return( tds_hd[i]);
    else return( tds_hd[tds_hd.size()-1]);
  }
  /** Returns Data time (sec) from the beginning */
  const double Time_elapsed(void) const {return time_elapsed;}
  /** Set Current time (legacy)*/
  void Set_Tcur(int d, int h, int m, double s){ 
    t_cur.Set_time(t_cur.Year(), d, h, m, s);}
  /** Set Current time (legacy)*/
  void Set_Tcur(Time_dat & t){ t_cur = t;}
  /** Returns current data time */
  const Time_dat & Time_current(void) const { return(t_cur);}
  /** Returns data time at beggining of data */
  const Time_dat & Time_bof(void) const {return t0;} //< Time at Beginning of file 
  /** Set flag to Skip Interpretation */
  void Set_NoInterpret(void){flag_NoInterpret=true;}
  /** Returns header array size*/
  int N_head()const {return tds_hd.size();}
private:
  /** Set internal buffer size for data rading */
  void Set_buf(const uint32_t buf_size);
  int Skip_process(void); ///<    Data Skip procedure 
  int Read2buf(void); ///<   Reading data to internal buffer 
  /** Set Initial data time */
  void Set_time_elapsed(double sec_t0){time_elapsed = sec_t0;}
  /** Add elapsed time */
  void Add_time_elapsed(double sec){time_elapsed += sec;}
  /** Reset rewind flag */
  void Reset_rewind_flag(void){flag_rewind=0;}
  /** Set to skip header byte */
  void Skip_hbyte(void){bp = data_buf.begin()+hbyte;}

  /*** Variables  */
  Time_dat t_cur;    /* initial time of Current data array  */
  double time_elapsed;         // seconds started from t0;
  FILE  *fp;
  string fname;
  vector <uint8_t> data_buf;  // data buffer 
  vector <TDS_HD>  tds_hd;
  //  uint32_t remain;  /* remaining data points No. in Buf to be read. */
  //  uint32_t  skip_byte; 
  float  skip_time;  // sec 
  uint32_t total_byte_sec; // Amount of byte/second excluding headder 
  uint32_t excpt_byte; // Skipping byte in Read2buf()
  Time_dat t0;         /// Initial epoch of the data in the file
  double t_span;       //  Time span of float data in sec.
  uint32_t ndata;      /// Number of data in this object. ndata = t_span*srate;
  vector <float> farr[4]; //  data array
  vector <float>::iterator  pfye[4];    /// Pointer to farray: end   */
  vector <double>  xarr;   /// second of year */
  vector <double>::iterator pdxe;    /// Pointer of xarr, start and end */
  char flag_rewind;    // Flag 1:rewind point for new data, 0:Initial 
  bool flag_NoInterpret;  // Skip iterprelation of data.
  int32_t n_extra;       /* extra data for farray in cross correlation */
  //  uint32_t fremain;  /* Remaining of data in float array from pfys*/
  //  int32_t Ofst_ini;   /* Initial offset corresponding to tau0 */
  int HD_Find(unsigned char  *);
  //  float* pFarr_tail(const int ich);
  const int Farr_Set(const int ch, const uint32_t npt, const float *pf){
    if(npt > 0 && (pfye[ch] + npt -1 < farr[ch].end()))  {
      memcpy(&pfye[ch][0],pf,npt*sizeof(float)); pfye[ch] += npt; return npt;}
    else  return -1;
  }
  vector <double>::iterator & pTarr_tail(void) {return pdxe;}
  //  virtual bool pTarr_inc(const int inc);
  // incremant farr iterator 
  bool pFarr_inc(const int ich, const int inc){
    if(pfye[ich] + inc-1 < farr[ich].end()){ pfye[ich]+=inc; return true; }
    else return false;
  }
};

//  inline float* TDS_FILE::pFarr_tail(const int ich){
//    if(ich<0 || ich>3){
//      fprintf(stderr, "Error pFarr(%d): index out of range\n",ich);
//      return 0;
//    }else     return &pfye[ich][0];
//  }

inline const int TDS_FILE::Update_info(const int level){
  if(tds_hd.size() <1){
    cerr << " Error in TDS_FILE::Update_info(): tds_hd is not set, yet\n";
    return -1;
  }
  TDS_HD & tdshd = Current_hd();
  
  double doy;
  if( tdshd.SyncB() == 0x8c)  // VSSP32
    doy = tdshd.Doy();
  else 
    doy = t_cur.Doy();

  double soy= (doy - 1)*86400 + tdshd.Sod();
  switch (level){
  case 0:   // default (all)
    //    t0.Set_time(t_cur.Year(), soy);
  case 1: // Update time, sampling parameters
    sfrq = tdshd.Sfrq();
    sbit = tdshd.Bit();
    nch = tdshd.Nch();
    // Pass to the next step 
  case 2: // update time,
    if( tdshd.SyncB() == 0x8c)  // VSSP32
      t_cur.Set_time(tdshd.Year(), soy);
    else
      t_cur.Set_time(t_cur.Year(), soy);
    // pass to the next step
  }
  return 0;
}


inline void TDS_FILE::Disp_hd(FILE *fp){
//   Time_dat t1;
//   vector <TDS_HD>::iterator phd = pr_tds_hd;
//   int doy = t_cur.Doy();
//   double sod = phd->Sod();
//   int h = (int)sod/3600; sod -= h*3600;
//   int m = (int)sod/60;  
//   double s = sod - m*60;
//   t1.Set_time(yydd[0], doy, h ,m, s);
  
  //  fprintf(fp, "---HD info -----------------------\n");
  //  t1.Print_time(fp,10);
  t_cur.Print_time(fp,10);
  fprintf(fp, " Nch=%d Sampl = %8d Bit=%d ", nch,sfrq,sbit );
  TDS_HD &phd = *(tds_hd.end()-1);
  phd.Print_inf(fp);
  if(phd.SyncB() == 0x8c){ // VSSP32
    fprintf(fp,"  FrmVer. %d.%d  Fmt# %d AUX-size:%d Fltr: %d MHz,", 
	    phd.MajVer(),phd.MinVer(),phd.FormatNum(),phd.AuxSize(),
	    phd.FilterFreq());
    fprintf(fp,"Stn:%s %s, Host:%s FrmErr:%d\n",
	    phd.StnID().c_str(), phd.Stname().c_str(), 
	    phd.Hostname().c_str(), phd.FlagFrameErr());
  }
}

inline void TDS_FILE::Set_buf(const uint32_t buf_size){ 
  data_buf.resize(buf_size); 
  total_byte_sec = buf_size-hbyte; 
  bp = data_buf.begin();
  bpe = bp;
}

inline const int32_t TDS_FILE::Read(const int msec ){
  uint32_t tmp_ndat = uint32_t(msec * 1.e-3 * Sfrq());
  if(ndata != tmp_ndat){
    ndata = tmp_ndat;
    Reset_rewind_flag();
  }
   for(int i=0; i<nch; i++){
    if(farr[i].size() !=ndata){
      farr[i].resize(ndata+n_extra);
      pfye[i] = farr[i].begin();
    }
  }
  if(xarr.size() != ndata){
    xarr.resize(ndata);
    pdxe = xarr.begin();
  }
  

  int32_t nget = Get_data(flag_rewind);
  flag_rewind=1;  
  if( nget < 0 ){
    cerr << "Error in TDS_FILE::Read(): returing from Get_data()\n";
    return -1;
  }

  if( flag_NoInterpret)    
    return total_byte_sec;
  else
    return nget;
}

#endif

/*
 * $Log: tds_samplr.hpp
 *  Author: $Author: sekido $
 *  Date: $Date: 2010/01/02 03:40:37 $
 *  $Header: /vsspfs1/work3/cvs-archive/VDIF/tds_file.hpp,v 1.2 2010/01/02 03:40:37 sekido Exp $
 *  Revision: $Revision: 1.2 $
 *  Source: $Source: /vsspfs1/work3/cvs-archive/VDIF/tds_file.hpp,v $
 */
